#include "stdafx.h"

#include "ff3se.h"
#include "monster.h"
#include "monsters.h"
#include "tile.h"

/*************************************/
//Calculates the amounf of room the Monster will take in the ROM
WORD Monster::calcSize(void)
{
	BYTE tileSize;
	BYTE numTiles = 0;
	int x, y;
	BYTE *ptr;

	if (numColors == 8)
		tileSize = 24;
	else
		tileSize = 32;

	if (size == 0)
	{
		ptr = &(monsters.molds8[mold * 64]);
		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
			{
				if (*(ptr++) == 1)
					numTiles++;
			}
		}
	}
	else
	{
		ptr = &(monsters.molds16[mold * 256]);
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
			{
				if (*(ptr++) == 1)
					numTiles++;
			}
		}
	}

	if ((tileSize * numTiles) == 0)
		return 8;
	else
		return tileSize * numTiles;
}

/*************************************/
//Changes the Monster's palette and updates the Bitmap pic
void Monster::changePalette(int p)
{
	int i;

	palette = p;

	for (i = 0; i < numColors; i++)
	{
		pic.setColor(i, monsters.monsPals[palette][i]);
	}
}

/*************************************/
//Clears the monster's image to a blank image
void Monster::clearImage(void)
{
	int i;

	if (size == 0)
		pic.createBlank(64, 64);
	else
		pic.createBlank(128, 128);

	for (i = 0; i < numColors; i++)
	{
		pic.setColor(i, monsters.monsPals[palette][i]);
	}

	getNewMold();
	createImage();
}

/*************************************/
//Copies the image used by Monster # num into the current image used by the monster
void Monster::copyImage(int num)
{
	BYTE max;
	int x, y;
	int img;
	int i;

	img = (monsters.getMonster(num))->imageNum;

	if (imageNum == img)						//Exit if already use that image
		return;

	mold = monsters.image[img].mold;
	numColors = monsters.image[img].numColors;
	size = monsters.image[img].size;

	if (size == 0)
		max = 64;
	else
		max = 128;

	pic.createBlank(max, max);				//Set Bitmap to new correct size

	copyPalette((monsters.getMonster(num))->palette);

	for (i = 0; i < numColors; i++)			//Update palette in Bitmap
		pic.setColor(i, monsters.monsPals[palette][i]);

	for (y = 0; y < max; y++)					//Copy the image from that monster
	{
		for (x = 0; x < max; x++)
		{
			pic.setPixel(x, y, (monsters.getMonster(num))->pic.getPixel(x, y));
		}
	}

	createImage();							//Update the image data based on the Bitmap
}

/*************************************/
//Copies palette #num into the palette currently used by the monster
void Monster::copyPalette(int num)
{
	BYTE size = numColors * 3;

	memcpy(&(monsters.monsPals[palette]), &(monsters.monsPals[num]), size);
}

/*************************************/
//Creates a new Bitmap for the monster from it's image data
void Monster::createBitmap(BYTE *ptr)
{
	int x, y, i;
	BYTE tile[64];
	BYTE *moldPtr;

	if (ptr == NULL)
		ptr = monsters.image[imageNum].graphicsData;

	mold = monsters.image[imageNum].mold;
	size = monsters.image[imageNum].size;
	numColors = monsters.image[imageNum].numColors;

	if (size == 0)
	{
		pic.createBlank(64, 64);

		for (i = 0; i < numColors; i++)
		{
			pic.setColor(i, monsters.monsPals[palette][i]);
		}

		moldPtr = &(monsters.molds8[mold * 64]);

		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
			{
				if (*(moldPtr++) == 1)
				{
					loadTile(ptr, tile, numColors);
					pic.copyTile(tile, x, y);
					if (numColors == 16)
						ptr += 32;
					else
						ptr += 24;
				}
			}
		}
	}
	else
	{
		pic.createBlank(128, 128);

		for (i = 0; i < numColors; i++)
		{
			pic.setColor(i, monsters.monsPals[palette][i]);
		}

		moldPtr = &(monsters.molds16[mold * 256]);

		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
			{
				if (*(moldPtr++) == 1)
				{
					loadTile(ptr, tile, numColors);
					pic.copyTile(tile, x, y);
					if (numColors == 16)
						ptr += 32;
					else
						ptr += 24;
				}
			}
		}

	}

}

/*************************************/
//Creates the image data compatible with the ROM for the monster 
void Monster::createImage(void)
{
	BYTE *ptr;
	int x, y;
	BYTE *moldPtr;

	monsters.image[imageNum].mold = mold;
	monsters.image[imageNum].ROMsize = calcSize();
	monsters.image[imageNum].numColors = numColors;
	monsters.image[imageNum].size = size;

	SAFE_DELETE(monsters.image[imageNum].graphicsData);

	monsters.image[imageNum].graphicsData = (BYTE *)malloc(monsters.image[imageNum].ROMsize);
	ZeroMemory(monsters.image[imageNum].graphicsData, monsters.image[imageNum].ROMsize);

	ptr = monsters.image[imageNum].graphicsData;

	if (size == 0)
	{
		moldPtr = &(monsters.molds8[mold * 64]);

		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
			{
				if (*(moldPtr++) == 1)
				{
					createTile(pic, x, y, ptr, numColors);
					if (numColors == 16)
						ptr += 32;
					else
						ptr += 24;
				}
			}
		}
	}
	else
	{
		moldPtr = &(monsters.molds16[mold * 256]);

		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
			{
				if (*(moldPtr++) == 1)
				{
					createTile(pic, x, y, ptr, numColors);
					if (numColors == 16)
						ptr += 32;
					else
						ptr += 24;

				}
			}
		}
	}
}

/*************************************/
//Flips the Bitmap
//If flip is 0, flips it horizontally, else flips if vetically
void Monster::flipImage(int flip)
{
	pic.flipImage(flip, true);
	getNewMold();
	createImage();
}

/*************************************/
//Gets the minimum X and Y dimension to hold the monster's pic
void Monster::getDimensions(int *maxX, int *maxY)
{
	int x, y;
	*maxX = 0;
	*maxY = 0;

	if (size == 0)
	{
		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
			{
				if (monsters.molds8[mold * 64 + x + y * 8] == 1)
				{
					if (y > *maxY)
						*maxY = y;
					if (x > *maxX)
						*maxX = x;
				}
			}
		}
	}
	else
	{
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
			{
				if (monsters.molds16[mold * 256 + x + y * 16] == 1)
				{
					if (y > *maxY)
						*maxY = y;
					if (x > *maxX)
						*maxX = x;
				}
			}
		}
	}

	(*maxX)++;
	(*maxY)++;

	*maxX *= 8;
	*maxY *= 8;
}

/*************************************/
//Calculates the Monster's mold data based on it's bitmap
//Changes to a new mold num if necessary
void Monster::getNewMold(void)
{
	int x2, y2;
	BYTE *newMold;
	int yVal;
	int i;
	int found = -1;

	if (size == 0)
	{
		newMold = (BYTE *)malloc(64);
		ZeroMemory(newMold, 64);
		for (y2 = 0; y2 < 64; y2++)
		{
			yVal = ((y2 / 8) * 8);
			for (x2 = 0; x2 < 64; x2++)
			{
				if (pic.getPixel(x2, y2) != 0)
				{
					newMold[(x2 / 8) + yVal] = 1;
				}
			}
		}

		if (numTiles(newMold, 8) == 0)
			newMold[0] = 1;

		for (i = 0; (i < monsters.numMolds8) && (found == -1); i++)
		{
			if (moldcmp(newMold, monsters.molds8 + i * 64, size))
				found = i;
		}
		if (found == -1)
		{
			monsters.molds8 = (BYTE *)realloc(monsters.molds8, (monsters.numMolds8 + 1) * 64);
			memcpy(monsters.molds8 + (monsters.numMolds8 * 64), newMold, 64);
			mold = monsters.numMolds8++;
		}
		else
		{
			mold = found;
		}
	}
	else
	{
		newMold = (BYTE *)malloc(256);
		ZeroMemory(newMold, 256);
		for (y2 = 0; y2 < 128; y2++)
		{
			yVal = ((y2 / 8) * 16);
			for (x2 = 0; x2 < 128; x2++)
			{
				if (pic.getPixel(x2, y2) != 0)
				{
					newMold[(x2 / 8) + yVal] = 1;
				}
			}
		}

		if (numTiles(newMold, 16) == 0)
			newMold[0] = 1;

		for (i = 0; (i < monsters.numMolds16) && (found == -1); i++)
		{
			if (moldcmp(newMold, monsters.molds16 + (i * 256), size))
				found = i;
		}
		if (found == -1)
		{
			monsters.molds16 = (BYTE *)realloc(monsters.molds16, (monsters.numMolds16 + 1) * 256);
			memcpy(monsters.molds16 + (monsters.numMolds16 * 256), newMold, 256);
			mold = monsters.numMolds16++;
		}
		else
		{
			mold = found;
		}
	}


	SAFE_DELETE(newMold);
}

/*************************************/
//Loads the monster's graphics from the ROM
void Monster::loadGraphics(void)
{
	BYTE *ptr;
	ptr = monsters.graphics;

	ptr += (offset * 8);

	//Copy graphics data to image
	SAFE_DELETE(monsters.image[imageNum].graphicsData);
	monsters.image[imageNum].graphicsData = (BYTE *)malloc(monsters.image[imageNum].ROMsize);
	memcpy(monsters.image[imageNum].graphicsData, ptr, monsters.image[imageNum].ROMsize);

	createBitmap(ptr);
}

/*************************************/
//Loads the monster's stats from it's 5 byte visual specification passed as ptr
void Monster::loadStats(BYTE *ptr)
{
	int found;
	int i;

	offset = *(ptr++);
	if (((*ptr) & 0x80) != 0)
	{
		numColors = 8;
		offset += (*(ptr++) & 0x7F) * 256;
	}
	else
	{
		numColors = 16;
		offset += (*(ptr++)) * 256;
	}
	if ((*(ptr) & 0x80) != 0)
	{
		size = 1;
	}
	else
	{
		size = 0;
	}

	palette = 0;
	if ((*(ptr) & 0x01) != 0)
		palette += 256;
	if ((*(ptr++) & 0x02) != 0)
		palette += 512;
	palette += *(ptr++);
	mold = *(ptr++);

	found = -1;

	i = 0;
	do {
		if (monsters.image[i].offset == offset)
			found = i;
	} while ((monsters.image[i++].offset != -1) && (found == -1));

	if (found != -1)
		i = found;
	else
		i--;

	monsters.image[i].mold = mold;
	monsters.image[i].offset = offset;
	monsters.image[i].ROMsize = calcSize();
	monsters.image[i].size = size;
	monsters.image[i].numColors = numColors;

	imageNum = i;

	if (imageNum > monsters.numMonsImages)
		monsters.numMonsImages = imageNum;
}


/*************************************/
//Creates a mew blank image for the monster
void Monster::newImage(void)
{
	int i;

	if (size == 0)
		pic.createBlank(64, 64);
	else
		pic.createBlank(128, 128);

	for (i = 0; i < numColors; i++)
	{
		pic.setColor(i, monsters.monsPals[palette][i]);
	}

	imageNum = monsters.numMonsImages++;

	getNewMold();
	createImage();
}

/*************************************/
//Saves the Monster's spec data to the ROM
void Monster::saveSpecs(int handle)
{
	BYTE specs[5];

	ZeroMemory(&specs, 5);
	specs[4] = mold;
	if (palette & 512)
		specs[2] += 2;
	if (palette & 256)
		specs[2] += 1;
	if (size == 1)
		specs[2] += 0x80;
	specs[3] = palette & 0xFF;
	specs[0] = monsters.image[imageNum].offset & 0xFF;
	specs[1] = (monsters.image[imageNum].offset & 0x7F00) >> 8;
	if (numColors == 8)
		specs[1] += 0x80;
	_write(handle, &specs, 5);
}

/*************************************/
//Copies the Bitmap passed as bmp into the monster's Bitmap
//Updates the image data
void Monster::setBitmap(Bitmap& bmp)
{
	int x, y, w, h;
	int i;

	w = 128;
	h = 128;

	if (size == 0)
	{
		if ((bmp.getHeight() > 64) || (bmp.getWidth() > 64))
			size = 1;
		else
		{
			w = 64;
			h = 64;
		}
	}
	else
	{
		if ((bmp.getHeight() <= 64) && (bmp.getWidth() <= 64))
		{
			size = 0;
			w = 64;
			h = 64;
		}
	}

	if (bmp.numColors() > 8)
		numColors = 16;
	else
		numColors = 8;

	for (i = 0; i < numColors; i++)
	{
		monsters.monsPals[palette][i].rgbtBlue = bmp.getColor(i).rgbtBlue;
		monsters.monsPals[palette][i].rgbtGreen = bmp.getColor(i).rgbtGreen;
		monsters.monsPals[palette][i].rgbtRed = bmp.getColor(i).rgbtRed;
	}

	pic.createBlank(w, h);

	for (i = 0; i < numColors; i++)
	{
		pic.setColor(i, monsters.monsPals[palette][i]);
	}

	for (y = 0; y < h; y++)
	{
		for (x = 0; x < w; x++)
		{
			if ((y >= bmp.getHeight()) || (x >= bmp.getWidth()))
				pic.setPixel(x, y, 0);
			else
				pic.setPixel(x, y, bmp.getPixel(x, y));
		}
	}

	changePalette(palette);		//Update the palette in the Monster's Bitmap

	//Write Bitmap data to image graphics data
	createImage();
}

/*************************************/
//Sets the monsters' image data to img
//Updates the monster's bitmap
void Monster::setImage(int img)
{
	imageNum = img;
	mold = monsters.image[imageNum].mold;
	size = monsters.image[imageNum].size;
	numColors = monsters.image[imageNum].numColors;

	createBitmap();
}

/*************************************/
//Changes the number of colors used by the Monster
void Monster::setNumColors(BYTE num)
{
	numColors = num;

	createImage();
	createBitmap();
}

/*************************************/
//Sets pixel x,y in the Monster's Bitmap to color c
//Alters the mold data if necessary
//Returns the monster's mold number
int Monster::setPixel(int x, int y, BYTE c)
{
	pic.setPixel(x, y, c);

	getNewMold();

	return mold;
}

/*************************************/
//Changes the size of the monster
void Monster::setSize(int newSize)
{
	size = newSize;

	getNewMold();

	createImage();
	createBitmap();
}

/*************************************/
//Makes color # num transparent by swapping it with color 0
void Monster::setTransparency(BYTE num)
{
	RGBTRIPLE temp;
	int x, y;

	//Make sure nonused tiles stay transparent
	if (size == 0)
	{
		for (y = 0; y < 64; y++)
		{
			for (x = 0; x < 64; x++)
			{
				if (monsters.molds8[mold * 64 + (x / 8) + (y / 8) * 8] == 0)
					pic.setPixel(x, y, num);
			}
		}
	}
	else
	{
		for (y = 0; y < 128; y++)
		{
			for (x = 0; x < 128; x++)
			{
				if (monsters.molds16[mold * 256 + (x / 8) + (y / 8) * 16] == 0)
					pic.setPixel(x, y, num);
			}
		}
	}

	pic.swapPal(0, num);

	temp = monsters.monsPals[palette][0];

	monsters.monsPals[palette][0].rgbtBlue = monsters.monsPals[palette][num].rgbtBlue;
	monsters.monsPals[palette][0].rgbtGreen = monsters.monsPals[palette][num].rgbtGreen;
	monsters.monsPals[palette][0].rgbtRed = monsters.monsPals[palette][num].rgbtRed;
	monsters.monsPals[palette][num].rgbtBlue = temp.rgbtBlue;
	monsters.monsPals[palette][num].rgbtGreen = temp.rgbtGreen;
	monsters.monsPals[palette][num].rgbtRed = temp.rgbtRed;
}

